/*	NetBSD: dotlock.c,v 1.11 2009/10/21 01:07:46 snj Exp	*/

/*
 * Copyright (c) 1996 Christos Zoulas.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "sh.h"

#include <stdio.h>
#ifndef O_SYNC
#define O_SYNC	0
#endif
#include <unistd.h>
#include <fcntl.h>
 
#include "dotlock.h"

static int create_exclusive(const char *);
/*
 * Create a unique file. O_EXCL does not really work over NFS so we follow
 * the following trick: [Inspired by  S.R. van den Berg]
 *
 * - make a mostly unique filename and try to create it.
 * - link the unique filename to our target
 * - get the link count of the target
 * - unlink the mostly unique filename
 * - if the link count was 2, then we are ok; else we've failed.
 */
static int
create_exclusive(const char *fname)
{
    	struct flock fl = { F_WRLCK, SEEK_SET, 0, 0, getpid()};
	int fd, retval;

	fd = open(fname, O_WRONLY|O_CREAT|O_SYNC, S_IWUSR);
	if (fd < 0)
		return -1;
	do {
		int ret;
		retval = fcntl(fd, F_SETLKW, &fl);
		if (retval < 0) {
			if (errno == EINTR)
				continue;
			close(fd);
			fd = -1;
			break;
		}
	} while (retval < 0);

	(void)unlink(fname);
	return fd;
}

/*
 * fname -- Pathname to lock
 * pollinterval -- Interval (miliseconds) to check for lock, -1 return
 */
int
dot_lock(const char *fname, int pollinterval)
{
	char path[MAXPATHLEN];
	sigset_t nset, oset;
	int retval;

	(void)sigemptyset(&nset);
	(void)sigaddset(&nset, SIGHUP);
	(void)sigaddset(&nset, SIGINT);
	(void)sigaddset(&nset, SIGQUIT);
	(void)sigaddset(&nset, SIGTERM);
	(void)sigaddset(&nset, SIGTTIN);
	(void)sigaddset(&nset, SIGTTOU);
	(void)sigaddset(&nset, SIGTSTP);
	(void)sigaddset(&nset, SIGCHLD);

	(void)snprintf(path, sizeof(path), "%s.lock", fname);

	retval = -1;
	handle_pending_signals();
	(void)sigprocmask(SIG_BLOCK, &nset, &oset);
	retval = create_exclusive(path);
	(void)sigprocmask(SIG_SETMASK, &oset, NULL);
	handle_pending_signals();
	return retval;
}

void
dot_unlock(const int fd)
{
    	struct flock fl = { F_UNLCK, SEEK_SET, 0, 0, getpid()};
	fcntl(fd, F_SETLK, &fl);
	close(fd);
}
